/*
 * $Id: WikiHandler.java,v 1.4 2003/07/13 10:48:22 kenta Exp $
 *
 * Copyright (C) 2003 Kenta Cho. All rights reserved.
 */
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.NoSuchElementException;

import javax.microedition.io.Connector;

import com.nttdocomo.io.HttpConnection;

/**
 * Access wiki with XML-RPC.
 */
public class WikiHandler {
    private String url;

    public void setURL(String u) {
        url = u;
    }

    /**
     * wiki.getPage
     * @param page WikiName
     * @return String Wiki page
     * @throws IOException failed to connect the wiki server.
     * @throws NoSuchElementException No such wiki page.
     */
    public String getPage(String page)
        throws IOException, NoSuchElementException {
        if (url == null || url.length() <= 0) {
            throw new NoSuchElementException("NO URL.");
        }
        try {
            String rs =
                callXmlRpc(
                    "<methodCall><methodName>wiki.getPage</methodName><params><param><value><string>"
                        + convertStringtoNCR(page)
                        + "</string></value></param></params></methodCall>");
            if (rs.indexOf("<fault>") != -1) {
                throw new NoSuchElementException("No such page");
            }
            int si = rs.indexOf("<base64>") + 8;
            int ei = rs.indexOf("</base64>", si);
            if (ei == -1) {
                ei = rs.indexOf("<", si);
                if (ei == -1) {
                    ei = rs.length();
                }
            }
            String r = rs.substring(si, ei);
            byte[] rb = Base64.decode(r);
            //return convertNCRtoString(getUTFString(rb));
            return getUTFString(rb);
        } catch (IOException e) {
            throw e;
        } catch (NoSuchElementException e) {
            throw e;
        } catch (Exception e) {
            throw new IOException("getPage failed: " + e.getMessage());
        }
    }

    /**
     * wiki.putPage
     * @param page WikiName
     * @param text Wiki page
     * @throws IOException failed to connect the server.
     */
    public void putPage(String page, String text) throws IOException {
        if (url == null || url.length() <= 0) {
            throw new IOException("NO URL.");
        }
        try {
            //String ps = Base64.encode(convertStringtoNCR(text).getBytes());
            String ps = Base64.encode(getUTFBytes(text));
            String rs =
                callXmlRpc(
                    "<methodCall><methodName>wiki.putPage</methodName><params><param><value><string>"
                        + convertStringtoNCR(page)
                        + "</string></value></param><param><value><base64>"
                        + ps
                        + "</base64></value></param></params></methodCall>");
            String r =
                rs.substring(
                    rs.indexOf("<boolean>") + 9,
                    rs.indexOf("</boolean>"));
            if (!r.equals("1")) {
                throw new IOException("wiki.putPage failed.");
            }
        } catch (IOException e) {
            throw e;
        } catch (Exception e) {
            throw new IOException("putPage failed: " + e.getMessage());
        }
    }

    private String callXmlRpc(String xml) throws IOException {
        HttpConnection cn = null;
        InputStream in = null;
        OutputStream out = null;
        byte[] rs;
        try {
            cn =
                (HttpConnection) Connector.open(
                    url,
                    Connector.READ_WRITE,
                    true);
            cn.setRequestMethod(HttpConnection.POST);
            cn.setRequestProperty("Content-Type", "text/xml");
            out = cn.openOutputStream();
            out.write(xml.getBytes());
            out.close();
            out = null;
            cn.connect();

            in = cn.openInputStream();
            rs = new byte[Page.MS];
            in.read(rs, 0, rs.length - 2);
        } finally {
            if (out != null)
                out.close();
            if (in != null)
                in.close();
            if (cn != null)
                cn.close();
        }
        return new String(rs);
    }

    public static String getUTFString(byte[] b, int ofs) throws IOException {
        byte[] nb = new byte[b.length + 2 - ofs];
        System.arraycopy(b, 0, nb, 2, b.length - ofs);
        nb[0] = (byte) (((b.length - ofs) & 0xff00) >> 8);
        nb[1] = (byte) ((b.length - ofs) & 0xff);
        ByteArrayInputStream bs = new ByteArrayInputStream(nb);
        DataInputStream ds = new DataInputStream(bs);
        String rs = ds.readUTF();
        ds.close();
        bs.close();
        return rs;
    }

    public static String getUTFString(byte[] b) throws IOException {
        String rsl = null;
        // Retry 6 times when the UTF bytes are terminated incorrectly.
        for (int i = 0; i < 6; i++) {
            try {
                rsl = getUTFString(b, i);
            } catch (IOException e) {}
            if (rsl != null)
                return rsl;
        }
        throw new IOException("failed to read UTF bytes");
    }

    public static byte[] getUTFBytes(String s) throws IOException {
        ByteArrayOutputStream bs = new ByteArrayOutputStream();
        DataOutputStream ds = new DataOutputStream(bs);
        ds.writeUTF(s);
        byte[] us = bs.toByteArray();
        byte[] rs = new byte[us.length - 2];
        System.arraycopy(us, 2, rs, 0, rs.length);
        ds.close();
        bs.close();
        return rs;
    }

	/*private String convertNCRtoString(String is) {
		StringBuffer rs = new StringBuffer();
		boolean amp = false;
		for (int i = 0;;) {
			if (i >= is.length())
				break;
			char c = is.charAt(i);
			if (amp) {
				amp = false;
				if (c == '#') {
					int ei = is.indexOf(';', i);
					int n = Integer.parseInt(is.substring(i + 1, ei));
					char ch = (char) n;
					rs.append(ch);
					i = ei + 1;
					continue;
				} else {
					rs.append('&');
				}
			} else {
				if (c == '&') {
					amp = true;
					i++;
					continue;
				}
			}
			rs.append(c);
			i++;
		}
		return rs.toString();
	}*/

    private String convertStringtoNCR(String in) {
        StringBuffer rs = new StringBuffer();
        for (int i = 0; i < in.length(); i++) {
            int n = (int) in.charAt(i);
            if (n < 0x80) {
                rs.append((char) n);
            } else {
                rs.append("&#" + n + ";");
            }
        }
        return rs.toString();
    }
}
